from dataclasses import dataclass
from typing import Collection, Sequence, Tuple

from .general_sam import Trie


def construct_trie_from_chars(
    strings: Collection[str],
) -> Tuple[Trie, Sequence[int]]:
    trie = Trie.in_chars()
    node_ids = [trie.insert_chars(s) for s in strings]
    return trie, node_ids


def construct_trie_from_bytes(
    strings: Collection[bytes],
) -> Tuple[Trie, Sequence[int]]:
    trie = Trie.in_bytes()
    node_ids = [trie.insert_bytes(s) for s in strings]
    return trie, node_ids


@dataclass
class CountInfo:
    str_cnt: int
    tot_cnt_lower: int
    tot_cnt_upper: int


@dataclass
class SortResult:
    trie: Trie
    node_ids: Sequence[int]
    cnt_info_on_nodes: Sequence[CountInfo]
    cnt_info_on_strings: Sequence[CountInfo]
    order: Sequence[int]
    rank: Sequence[int]


def sort_chars(strings: Collection[str]) -> SortResult:
    trie, node_ids = construct_trie_from_chars(strings)
    return sort_seq_via_trie(trie, node_ids)


def sort_bytes(strings: Collection[bytes]) -> SortResult:
    trie, node_ids = construct_trie_from_bytes(strings)
    return sort_seq_via_trie(trie, node_ids)


def sort_seq_via_trie(trie: Trie, node_ids: Sequence[int]) -> SortResult:
    num_of_seq = len(node_ids)

    cnt_info_on_nodes = [CountInfo(0, 0, 0) for _ in range(trie.num_of_nodes())]
    for k in node_ids:
        cnt_info_on_nodes[k].str_cnt += 1

    tot_str_cnt = 0

    def in_stack(node_id: int, _):
        nonlocal tot_str_cnt
        node_info = cnt_info_on_nodes[node_id]
        node_info.tot_cnt_lower = tot_str_cnt
        tot_str_cnt += node_info.str_cnt

    def out_stack(node_id: int):
        nonlocal tot_str_cnt
        node_info = cnt_info_on_nodes[node_id]
        node_info.tot_cnt_upper = tot_str_cnt

    trie.dfs_travel(in_stack, out_stack)

    cnt_info_on_strings = [cnt_info_on_nodes[node_ids[i]] for i in range(num_of_seq)]

    order = sorted(
        range(num_of_seq),
        key=lambda i: cnt_info_on_strings[i].tot_cnt_lower,
    )
    rank = [0] * num_of_seq
    for k, i in enumerate(order):
        rank[i] = k

    return SortResult(
        trie=trie,
        node_ids=node_ids,
        cnt_info_on_nodes=cnt_info_on_nodes,
        cnt_info_on_strings=cnt_info_on_strings,
        order=order,
        rank=rank,
    )



from .general_sam import (
    GeneralSAM,
    GeneralSAMState,
    Trie,
    TrieNode,
)
from .trie_utils import (
    CountInfo,
    SortResult,
    construct_trie_from_bytes,
    construct_trie_from_chars,
    sort_bytes,
    sort_chars,
    sort_seq_via_trie,
)
from .vocab_prefix import (
    VocabPrefixAutomaton,
    VocabPrefixBytesOrChars,
)

__all__ = [
    'GeneralSAM',
    'GeneralSAMState',
    'Trie',
    'TrieNode',
    'CountInfo',
    'SortResult',
    'construct_trie_from_chars',
    'construct_trie_from_bytes',
    'sort_chars',
    'sort_bytes',
    'sort_seq_via_trie',
    'VocabPrefixAutomaton',
    'VocabPrefixBytesOrChars',
]



import enum
from dataclasses import replace
from typing import (
    Callable,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
    cast,
)

from .general_sam import GeneralSAM, GeneralSAMState, Trie
from .trie_utils import (
    CountInfo,
    SortResult,
    construct_trie_from_bytes,
    construct_trie_from_chars,
    sort_bytes,
    sort_chars,
)


class VocabPrefixBytesOrChars(enum.Enum):
    BYTES = enum.auto()
    CHARS = enum.auto()


class VocabPrefixAutomaton(object):
    def __init__(
        self,
        vocab: Iterable[Union[str, bytes]],
        bytes_or_chars: Union[
            str, VocabPrefixBytesOrChars
        ] = VocabPrefixBytesOrChars.CHARS,
    ) -> None:
        if isinstance(bytes_or_chars, str):
            bytes_or_chars = getattr(VocabPrefixBytesOrChars, bytes_or_chars.upper())

        self.bytes_or_chars = cast(VocabPrefixBytesOrChars, bytes_or_chars)

        self.vocab: Sequence[Union[str, bytes]] = list(vocab)

        if self.bytes_or_chars == VocabPrefixBytesOrChars.BYTES and isinstance(
            self.vocab[0], str
        ):
            self.vocab = list(cast(str, i).encode() for i in self.vocab)
        if self.bytes_or_chars == VocabPrefixBytesOrChars.CHARS and isinstance(
            self.vocab[0], bytes
        ):
            self.vocab = list(cast(bytes, i).decode() for i in self.vocab)

        self.vocab_rev: Sequence[Union[str, bytes]] = list(s[::-1] for s in vocab)

        sort_seq, construct_trie = {
            VocabPrefixBytesOrChars.BYTES: (sort_bytes, construct_trie_from_bytes),
            VocabPrefixBytesOrChars.CHARS: (sort_chars, construct_trie_from_chars),
        }[self.bytes_or_chars]
        self.vocab_sort_res = cast(SortResult, sort_seq(self.vocab))
        self.trie_rev, self.trie_rev_node_ids = cast(
            Tuple[Trie, Sequence[int]],
            construct_trie(self.vocab_rev),
        )

        self.sam_rev = GeneralSAM.construct_from_trie(self.trie_rev)
        self._gen_cnt_info_in_sam()

    @property
    def _state_feed_fn(self) -> Callable[[GeneralSAMState, Union[bytes, str]], None]:
        return {
            VocabPrefixBytesOrChars.BYTES: GeneralSAMState.feed_bytes,
            VocabPrefixBytesOrChars.CHARS: GeneralSAMState.feed_chars,
        }[self.bytes_or_chars]

    def _gen_cnt_info_in_sam(self):
        self.cnt_info_in_sam: List[Optional[CountInfo]] = [
            None for _ in range(self.sam_rev.num_of_nodes())
        ]

        for token_rev, cnt_info in zip(
            self.vocab_rev, self.vocab_sort_res.cnt_info_on_strings
        ):
            state = self.sam_rev.get_root_state()
            self._state_feed_fn(state, token_rev)
            state_id = state.get_node_id()
            self.cnt_info_in_sam[state_id] = replace(cnt_info, str_cnt=1)

        for sam_state in reversed(self.sam_rev.get_topo_order()):
            assert not sam_state.is_nil()
            if sam_state.is_root():
                continue

            state_id = sam_state.get_node_id()
            state_cnt_info = self.cnt_info_in_sam[state_id]
            if state_cnt_info is None:
                continue

            link_id = sam_state.get_suffix_parent_id()
            link_cnt_info = self.cnt_info_in_sam[link_id]

            if link_cnt_info is None:
                self.cnt_info_in_sam[link_id] = replace(state_cnt_info)
                continue

            link_cnt_info.str_cnt += state_cnt_info.str_cnt
            link_cnt_info.tot_cnt_lower = min(
                link_cnt_info.tot_cnt_lower,
                state_cnt_info.tot_cnt_lower,
            )
            link_cnt_info.tot_cnt_upper = max(
                link_cnt_info.tot_cnt_upper,
                state_cnt_info.tot_cnt_upper,
            )

        for state_id in range(self.sam_rev.num_of_nodes()):
            sam_state = self.sam_rev.get_state(state_id)
            state_cnt_info = self.cnt_info_in_sam[state_id]
            if sam_state.is_nil() or sam_state.is_root() or state_cnt_info is None:
                continue

            link_id = sam_state.get_suffix_parent_id()
            link_cnt_info = self.cnt_info_in_sam[link_id]

            assert link_cnt_info is not None
            assert link_cnt_info.tot_cnt_lower <= state_cnt_info.tot_cnt_lower
            assert link_cnt_info.tot_cnt_upper >= state_cnt_info.tot_cnt_upper

    def get_root_state(self) -> GeneralSAMState:
        return self.sam_rev.get_root_state()

    def prepend_feed(
        self, state: GeneralSAMState, token: Union[str, bytes]
    ) -> Optional[CountInfo]:
        if self.bytes_or_chars == VocabPrefixBytesOrChars.BYTES and isinstance(
            token, str
        ):
            token = token.encode()
        self._state_feed_fn(state, token[::-1])
        return self.cnt_info_in_sam[state.get_node_id()]

    def get_order(self) -> Sequence[int]:
        return self.vocab_sort_res.order

    def get_order_slice(self, cnt_info: CountInfo) -> Sequence[int]:
        return self.vocab_sort_res.order[
            cnt_info.tot_cnt_lower : cnt_info.tot_cnt_upper
        ]



from typing import Callable, Mapping, Optional, Sequence, Union


class TrieNode:
    def is_in_chars(self) -> bool:
        ...

    def is_in_bytes(self) -> bool:
        ...

    def get_node_id(self) -> int:
        ...

    def is_accepting(self) -> bool:
        ...

    def get_trans(self) -> Mapping[Union[str, int], int]:
        ...

    def get_parent(self) -> int:
        ...


class Trie:
    @staticmethod
    def in_chars() -> 'Trie':
        ...

    @staticmethod
    def in_bytes() -> 'Trie':
        ...

    def is_in_chars(self) -> bool:
        ...

    def is_in_bytes(self) -> bool:
        ...

    def num_of_nodes(self) -> int:
        ...

    def insert_chars(self, s: str) -> int:
        ...

    def insert_bytes(self, s: bytes) -> int:
        ...

    def get_bfs_order(self) -> Sequence[int]:
        ...

    def get_root(self) -> TrieNode:
        ...

    def get_node(self, node_id: int) -> Optional[TrieNode]:
        ...

    def dfs_travel(
        self,
        in_stack_callback: Callable[[int, Optional[str]], None],
        out_stack_callback: Callable[[int], None],
        root_node_id: Optional[int] = None,
    ) -> TrieNode:
        ...

    def bfs_travel(
        self,
        in_queue_callback: Callable[[int, Optional[str]], None],
        out_queue_callback: Callable[[int], None],
        root_node_id: Optional[int] = None,
    ) -> TrieNode:
        ...


class GeneralSAMState:
    def is_in_str(self) -> bool:
        ...

    def is_in_bytes(self) -> bool:
        ...

    def get_node_id(self) -> int:
        ...

    def is_nil(self) -> bool:
        ...

    def is_root(self) -> bool:
        ...

    def is_accepting(self) -> bool:
        ...

    def get_trans(self) -> Mapping[Union[str, int], int]:
        ...

    def get_suffix_parent_id(self) -> int:
        ...

    def copy(self) -> 'GeneralSAMState':
        ...

    def goto_suffix_parent(self):
        ...

    def goto_char(self, t: str):
        ...

    def goto_byte(self, t: int):
        ...

    def feed_chars(self, s: str):
        ...

    def feed_bytes(self, s: bytes):
        ...

    def dfs_along(
        self,
        trie: Trie,
        in_stack_callback: Callable[['GeneralSAMState', int, Optional[str]], None],
        out_stack_callback: Callable[['GeneralSAMState', int], None],
        trie_node_id: Optional[int] = None,
    ) -> TrieNode:
        ...

    def bfs_along(
        self,
        trie: Trie,
        in_queue_callback: Callable[['GeneralSAMState', int, Optional[str]], None],
        out_queue_callback: Callable[['GeneralSAMState', int], None],
        trie_node_id: Optional[int] = None,
    ) -> TrieNode:
        ...


class GeneralSAM:
    @staticmethod
    def construct_from_chars(s: str) -> 'GeneralSAM':
        ...

    @staticmethod
    def construct_from_bytes(s: bytes) -> 'GeneralSAM':
        ...

    @staticmethod
    def construct_from_trie(trie: Trie) -> 'GeneralSAM':
        ...

    def is_in_str(self) -> bool:
        ...

    def is_in_bytes(self) -> bool:
        ...

    def num_of_nodes(self) -> int:
        ...

    def get_root_state(self) -> GeneralSAMState:
        ...

    def get_state(self, node_id: int) -> GeneralSAMState:
        ...

    def get_topo_order(self) -> Sequence[GeneralSAMState]:
        ...



from general_sam import VocabPrefixAutomaton, CountInfo


def test_chinese_chars_vocab_prefix():
    vocab = ['歌曲', '聆听歌曲', '播放歌曲', '歌词', '查看歌词']
    automaton = VocabPrefixAutomaton(vocab, bytes_or_chars='chars')

    # NOTE: CountInfo is related to the sorted vocab:
    _ = ['播放歌曲', '查看歌词', '歌曲', '歌词', '聆听歌曲']

    # 一起 | 聆 | 听 | 歌
    state = automaton.get_root_state()

    # feed 歌
    cnt_info = automaton.prepend_feed(state, '歌')
    assert cnt_info is not None and cnt_info == CountInfo(
        str_cnt=2, tot_cnt_lower=2, tot_cnt_upper=4
    )

    selected_idx = automaton.get_order_slice(cnt_info)
    assert frozenset(selected_idx) == {0, 3}
    selected_vocab = [vocab[i] for i in selected_idx]
    assert frozenset(selected_vocab) == {'歌曲', '歌词'}

    # feed 听
    cnt_info = automaton.prepend_feed(state, '听')
    assert cnt_info is None
    assert not state.is_nil()

    # feed 聆
    cnt_info = automaton.prepend_feed(state, '聆')
    assert cnt_info is not None and cnt_info == CountInfo(
        str_cnt=1, tot_cnt_lower=4, tot_cnt_upper=5
    )

    selected_idx = automaton.get_order_slice(cnt_info)
    assert frozenset(selected_idx) == {1}
    selected_vocab = [vocab[i] for i in selected_idx]
    assert frozenset(selected_vocab) == {'聆听歌曲'}

    # feed 一起
    assert not state.is_nil()
    cnt_info = automaton.prepend_feed(state, '一起')
    assert state.is_nil()

    # 来 | 查看 | 歌词
    state = automaton.get_root_state()

    # feed 歌词
    cnt_info = automaton.prepend_feed(state, '歌词')
    assert cnt_info is not None and cnt_info == CountInfo(
        str_cnt=1, tot_cnt_lower=3, tot_cnt_upper=4
    )

    selected_idx = automaton.get_order_slice(cnt_info)
    assert frozenset(selected_idx) == {3}
    selected_vocab = [vocab[i] for i in selected_idx]
    assert frozenset(selected_vocab) == {'歌词'}

    # feed 查看
    cnt_info = automaton.prepend_feed(state, '查看')
    assert cnt_info is not None and cnt_info == CountInfo(
        str_cnt=1, tot_cnt_lower=1, tot_cnt_upper=2
    )

    selected_idx = automaton.get_order_slice(cnt_info)
    assert frozenset(selected_idx) == {4}
    selected_vocab = [vocab[i] for i in selected_idx]
    assert frozenset(selected_vocab) == {'查看歌词'}

    # feed 来
    assert not state.is_nil()
    cnt_info = automaton.prepend_feed(state, '来')
    assert state.is_nil()



from general_sam import GeneralSAM, GeneralSAMState, construct_trie_from_chars


def test_bytes_abcbc():
    sam = GeneralSAM.construct_from_bytes(b'abcbc')

    state = sam.get_root_state()
    state.feed_bytes(b'cbc')
    assert state.is_accepting()

    state = sam.get_root_state()
    state.feed_bytes(b'bcb')
    assert not state.is_accepting()


def test_chars_abcbc():
    sam = GeneralSAM.construct_from_chars('abcbc')
    state = sam.get_root_state()

    state.feed_chars('b')
    assert not state.is_accepting()
    state.feed_chars('c')
    assert state.is_accepting()
    state.feed_chars('bc')
    assert state.is_accepting()
    state.feed_chars('bc')
    assert not state.is_accepting() and state.is_nil()


def test_simple_sam_from_trie():
    trie, _ = construct_trie_from_chars(['hello', 'Chielo'])
    sam = GeneralSAM.construct_from_trie(trie)

    def fetch_state(s: str) -> GeneralSAMState:
        state = sam.get_root_state()
        state.feed_chars(s)
        return state

    assert fetch_state('lo').is_accepting()
    assert fetch_state('ello').is_accepting()
    assert fetch_state('elo').is_accepting()

    state = fetch_state('el')
    assert not state.is_accepting() and not state.is_nil()

    state = fetch_state('bye')
    assert not state.is_accepting() and state.is_nil()
